#include <cstdio>
#include <iostream>
#include "circularbuffertracker.h"
#include "inputstage.h"
#include "processingstage.h"
#include "ioconfig.h"
#include "processingconfig.h"
//#include "scalarutils.h"
#include "outputstage.h"


template <typename T>
void runStages (const ProcessingConfig & processingConfig,
            const IOConfig         & ioConfig) {
    OutputStage *os = new OutputStage (ioConfig, NULL);
    ProcessingStage<T> *ps = new ProcessingStage<T> (processingConfig, os);
    InputStage<T> *is = new InputStage<T> (processingConfig, ioConfig, ps);
    os->start ();
    ps->start ();
    is->start ();

    is->wait ();
    ps->wait ();
    os->wait ();
}

int main (int argc, char *argv[]) {
    if ((3 > argc) || (5 < argc)){
        std::cerr << "Usage: " << argv[0] << " <input directory> <output directory> [BG ref frames=200 [FG sigma=0.02]]\n"
                     "Note that any contents of the output directory will be DELETED!" << std::endl;
        return EXIT_FAILURE;
    }




    IOConfig ioConfig;
    ProcessingConfig processingConfig;


    processingConfig.colourSpace = ProcessingConfig::RnGnGradS;
    processingConfig.colourDepth = 16;
//    processingConfig.blockWidth = processingConfig.blockHeight = 16;
    processingConfig.blockWidth = processingConfig.blockHeight = 8;
    processingConfig.roiAnalysis = true;
    processingConfig.spatialMarginalConditioning = true;
    processingConfig.morphologicalFilter = true;
    processingConfig.backgroundReferenceFrames = 200;
//    processingConfig.backgroundAppearanceSigma = 0.005;
    processingConfig.backgroundAppearanceSigma = 0.05;
    processingConfig.backgroundAppearanceSigmaEstimation = true;
    processingConfig.backgroundAppearanceSigmaEstimationMinimum = 0.001;
    processingConfig.backgroundAppearanceSigmaEstimationMaximum = 1.0;
    processingConfig.backgroundAppearanceSigmaEstimationDelay = 25;
    processingConfig.backgroundSpatialModelling = true;
    processingConfig.backgroundSpatialSigma = 0.471404521f;
    processingConfig.backgroundInverseSpatialSigma = 1.0 / processingConfig.backgroundSpatialSigma;
    processingConfig.backgroundSpatialRadius = static_cast<int>(floor(3.0 * processingConfig.backgroundSpatialSigma));
    processingConfig.backgroundSelectiveUpdate = true;
    processingConfig.backgroundSelectiveUpdateCutoff = 50;
    const float sigmaT = (processingConfig.backgroundReferenceFrames
                          - processingConfig.backgroundSelectiveUpdateCutoff) / 3.0f;
    processingConfig.backgroundSelectiveUpdateSigmaTsquare = sigmaT * sigmaT;

    processingConfig.foregroundModelEnabled = true;
    processingConfig.foregroundReferenceFrames = 10;
    processingConfig.foregroundAppearanceSigma = 0.02;
    processingConfig.foregroundUseBGAppearanceSigma = false;
//    processingConfig.foregroundUseBGAppearanceSigma = true;
//    processingConfig.foregroundBGAppearanceSigmaMultiplier = 1.0f;
    processingConfig.foregroundStartDelay = 25;
    processingConfig.foregroundSpatialModelling = true; // FG always features spatial modelling, right?
    processingConfig.foregroundSpatialSigma = 0.67f;
    processingConfig.foregroundSpatialRadius = static_cast<int>(floor(3.0 * processingConfig.foregroundSpatialSigma));
    processingConfig.foregroundMixingFactor = 0.01;

    processingConfig.particleFilterEnabled = true;
    processingConfig.particleFilterMaxRegions = 50;
    processingConfig.particleFilterMinRegionSize = 20;
    processingConfig.particleFilterMaxParticles = 1000;
    processingConfig.particleFilterMinParticleRadius = 2.5f;
    processingConfig.particleFilterMaxParticleRadius = 100.0f;
    processingConfig.particleFilterParticleLikelihoodBeta = 0.2;
    processingConfig.particleFilterRandomCentreJitter = 2.5f;
    processingConfig.particleFilterRandomSizeFactor = 2.0f;
    processingConfig.particleFilterComputeFGPrior = true;



    if (3 < argc) {
        processingConfig.backgroundReferenceFrames = std::atol(argv[3]);
        std::cerr << processingConfig.backgroundReferenceFrames << std::endl;
    }
    if (4 < argc) {
        processingConfig.foregroundAppearanceSigma = std::atof(argv[4]);
        std::cerr << processingConfig.foregroundAppearanceSigma << std::endl;
    }

    ioConfig.setInputDir(argv[1]);
    ioConfig.setOutputDir(argv[2]);
    ioConfig.setFilenameTemplate("fwefwfewrame-%04d.png");
    ioConfig.setFirstFrameNumber(1345345);
    ioConfig.setLastFrameNumber(224574575);
    ioConfig.setSavePNG(true);
    processingConfig.referenceBufferSize = 10 + std::max(processingConfig.backgroundReferenceFrames, processingConfig.foregroundReferenceFrames);
    ioConfig.outputBufferSize = 10;


//    std::cout << ioConfig;
//    std::cout << processingConfig;

    switch (processingConfig.colourDepth) {
    case 8:
        runStages<unsigned char> (processingConfig, ioConfig);
        break;
    case 16:
        runStages<unsigned short> (processingConfig, ioConfig);
        break;
    default:
        qFatal("Unsupported colour depth: %d\n", processingConfig.colourDepth);
    }
//    std::cout << ioConfig;
//    std::cout << processingConfig;
    return 0;
}
