#include "processingdatatypes.h"
#include "compiletimeconfig.h"



template <typename T>
ProcessingDeviceData<T>::ProcessingDeviceData()
    : //imageAbsDiffs(NULL),
      bgProbabilities(NULL),
//      imageSquaredDiffTemporalWeightedSum(NULL),
//      probTemporalSum(NULL),
      binsHistory(NULL),
      diffsHistograms(NULL),
      medianSigmas(NULL),
      backgroundAppearanceSigmas(NULL),
      probSpatialSum(NULL),
      numberOfFGSamples(NULL),
      pdfFG(NULL),
      pdfBG(NULL),
      morphologicalFG(NULL),
      fgMasks(NULL),
      foregroundReferencePixels(NULL),
      fgPriorProbability(NULL)
{}

template <typename T>
ProcessingDeviceData<T>::~ProcessingDeviceData() {
    this->releaseDeviceMemory();
}

template <typename T>
void ProcessingDeviceData<T>::allocateDeviceMemory(unsigned int width, unsigned int height, unsigned int channels, unsigned int numImages, unsigned int numForegroundReferenceFrames) {
//    assert(NULL == this->imageAbsDiffs);
    assert(NULL == this->bgProbabilities);
//    assert(NULL == this->imageSquaredDiffTemporalWeightedSum);
//    assert(NULL == this->probTemporalSum);
    assert(NULL == this->binsHistory);
    assert(NULL == this->diffsHistograms);
    assert(NULL == this->medianSigmas);
    assert(NULL == this->backgroundAppearanceSigmas);
    assert(NULL == this->probSpatialSum);
    assert(NULL == this->numberOfFGSamples);
    assert(NULL == this->pdfFG);
    assert(NULL == this->pdfBG);
    assert(NULL == this->morphologicalFG);
    assert(NULL == this->fgMasks);
    assert(NULL == this->foregroundReferencePixels);
    assert(NULL == this->fgPriorProbability);
//    CUDA_SAFE_CALL (cudaMalloc (&this->imageSquaredDiffTemporalWeightedSum, width
//                                * height
//                                * channels
//                                * sizeof (long)));
//    CUDA_SAFE_CALL (cudaMemset (this->imageSquaredDiffTemporalWeightedSum, 0, width
//                                * height
//                                * channels
//                                * sizeof (long)));
//    CUDA_SAFE_CALL (cudaMalloc (&this->probTemporalSum,  width
//                                * height
//                                * sizeof (long)));
//    CUDA_SAFE_CALL (cudaMemset (this->probTemporalSum, 0, width
//                                * height
//                                * sizeof (long)));
    CUDA_SAFE_CALL (cudaMalloc (&this->binsHistory, width * height * channels * numImages * sizeof (unsigned char)));
    CUDA_SAFE_CALL (cudaMalloc (&this->diffsHistograms, width * height * NUMBINS * channels * sizeof (unsigned short)));
    CUDA_SAFE_CALL (cudaMemset (this->diffsHistograms, 0, width * height * NUMBINS * channels * sizeof (unsigned short)));
    CUDA_SAFE_CALL (cudaMalloc (&this->medianSigmas, width * height * channels * sizeof (float)));
    CUDA_SAFE_CALL (cudaMalloc (&this->backgroundAppearanceSigmas, width * height * channels * sizeof (float)));
    CUDA_SAFE_CALL (cudaMalloc (&this->bgProbabilities, width
                                * height
                                * numImages
                                * sizeof (unsigned char)));
//    CUDA_SAFE_CALL (cudaMalloc (&this->imageAbsDiffs, width
//                                * height
//                                * channels
//                                * numImages
//                                * sizeof (T)));
    CUDA_SAFE_CALL (cudaMalloc (&this->probSpatialSum, sizeof (int) * numImages));
    CUDA_SAFE_CALL (cudaMalloc (&this->numberOfFGSamples, sizeof (int) * numImages));
    CUDA_SAFE_CALL (cudaMalloc (&this->pdfFG, width
                                * height
                                * sizeof (float)));
    CUDA_SAFE_CALL (cudaMalloc (&this->pdfBG, width
                                * height
                                * sizeof (float)));

    /// Layered buffer for post-processing FG detection. The operations
    /// to be performed on it are:
    /// - Closing: dilation + erosion
    /// - Opening: erosion + dilation
    /// - Erosion
    /// Each of the elemental operations must be done separately, therefore
    /// the buffer will have five layers plus one for the original data.
    CUDA_SAFE_CALL (cudaMalloc (&this->morphologicalFG, width
                                * height
                                * 6
                                * sizeof (unsigned char)));
    CUDA_SAFE_CALL (cudaMalloc (&this->fgMasks, width * height * numImages * sizeof(unsigned char)));
    CUDA_SAFE_CALL (cudaMalloc (&this->foregroundReferencePixels, width * height * numForegroundReferenceFrames * sizeof (ColouredPixel<T, 3>)));
    CUDA_SAFE_CALL (cudaMalloc(&this->fgPriorProbability, width * height * sizeof (float)));
}

template <typename T>
void ProcessingDeviceData<T>::releaseDeviceMemory() {
//    CUDA_SAFE_CALL(cudaFree(this->imageAbsDiffs));
//    this->imageAbsDiffs = NULL;
    CUDA_SAFE_CALL(cudaFree(this->bgProbabilities));
    this->bgProbabilities = NULL;
//    CUDA_SAFE_CALL(cudaFree(this->imageSquaredDiffTemporalWeightedSum));
//    this->imageSquaredDiffTemporalWeightedSum = NULL;
//    CUDA_SAFE_CALL(cudaFree(this->probTemporalSum));
//    this->probTemporalSum = NULL;
    CUDA_SAFE_CALL(cudaFree(this->binsHistory));
    this->binsHistory = NULL;
    CUDA_SAFE_CALL(cudaFree(this->diffsHistograms));
    this->diffsHistograms = NULL;
    CUDA_SAFE_CALL(cudaFree(this->medianSigmas));
    this->medianSigmas = NULL;
    CUDA_SAFE_CALL(cudaFree(this->backgroundAppearanceSigmas));
    this->backgroundAppearanceSigmas = NULL;
    CUDA_SAFE_CALL(cudaFree(this->probSpatialSum));
    this->probSpatialSum = NULL;
    CUDA_SAFE_CALL(cudaFree(this->numberOfFGSamples));
    this->numberOfFGSamples = NULL;
    CUDA_SAFE_CALL(cudaFree(this->pdfFG));
    this->pdfFG = NULL;
    CUDA_SAFE_CALL(cudaFree(this->pdfBG));
    this->pdfBG = NULL;
    CUDA_SAFE_CALL(cudaFree(this->morphologicalFG));
    this->morphologicalFG = NULL;
    CUDA_SAFE_CALL(cudaFree(this->fgMasks));
    this->fgMasks = NULL;
    CUDA_SAFE_CALL(cudaFree(this->foregroundReferencePixels));
    this->foregroundReferencePixels = NULL;
    CUDA_SAFE_CALL(cudaFree(this->fgPriorProbability));
    this->fgPriorProbability = NULL;
}

template struct ProcessingDeviceData<unsigned char>;
template struct ProcessingDeviceData<unsigned short>;


ProcessingDeviceAux::~ProcessingDeviceAux() {
    CUDA_SAFE_CALL(cudaFree(this->backgroundSpatialHorizontalGaussian));
    this->backgroundSpatialHorizontalGaussian = NULL;
    CUDA_SAFE_CALL(cudaFree(this->backgroundSpatialVerticalGaussian));
    this->backgroundSpatialVerticalGaussian = NULL;
    CUDA_SAFE_CALL(cudaFree(this->foregroundSpatialHorizontalGaussian));
    this->foregroundSpatialHorizontalGaussian = NULL;
    CUDA_SAFE_CALL(cudaFree(this->foregroundSpatialVerticalGaussian));
    this->foregroundSpatialVerticalGaussian = NULL;
}
